/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__bind_h
#define nonstd__bind_h 1

#include "tuple.h"
#include "function.h"

namespace nonstd
{
  namespace impl
  {
    template<unsigned int N> struct placeholder {};
  }


  namespace
  {
    impl::placeholder<0> _0;
    impl::placeholder<1> _1;
    impl::placeholder<2> _2;
    impl::placeholder<3> _3;
  }


  namespace impl
  {
    template<typename T, typename T0, typename T1, typename T2, typename T3>
    inline T
    select_value (const T &t, const tuple<T0, T1, T2, T3> &)
    {
      return t;
    }

    template<typename T0, typename T1, typename T2, typename T3>
    inline T0
    select_value (const placeholder<0> &, const tuple<T0, T1, T2, T3> &t)
    {
      return nth_elt<0> (t);
    }

    template<typename T0, typename T1, typename T2, typename T3>
    inline T1
    select_value (const placeholder<1> &, const tuple<T0, T1, T2, T3> &t)
    {
      return nth_elt<1> (t);
    }

    template<typename T0, typename T1, typename T2, typename T3>
    inline T2
    select_value (const placeholder<2> &, const tuple<T0, T1, T2, T3> &t)
    {
      return nth_elt<2> (t);
    }

    template<typename T0, typename T1, typename T2, typename T3>
    inline T3
    select_value (const placeholder<3> &, const tuple<T0, T1, T2, T3> &t)
    {
      return nth_elt<3> (t);
    }


    template<typename T> struct rank;

    template<>
    struct rank<nil>
    {
      static const unsigned int value = 0;
    };

    template<typename H, typename T>
    struct rank<cons<H, T> >
    {
      static const unsigned int value = rank<T>::value;
    };

    template <typename T>
    struct rank<cons<placeholder<0>, T> >
    {
      static const unsigned int value =
        rank<T>::value > 1 ? rank<T>::value : 1;
    };

    template <typename T>
    struct rank<cons<placeholder<1>, T> >
    {
      static const unsigned int value =
        rank<T>::value > 2 ? rank<T>::value : 2;
    };

    template <typename T>
    struct rank<cons<placeholder<2>, T> >
    {
      static const unsigned int value =
        rank<T>::value > 3 ? rank<T>::value : 3;
    };

    template <typename T>
    struct rank<cons<placeholder<3>, T> >
    {
      static const unsigned int value =
        rank<T>::value > 4 ? rank<T>::value : 4;
    };


    template<unsigned int N> struct binder_call;

    template<>
    struct binder_call<0>
    {
      template<typename Fn,
               typename P0, typename P1, typename P2, typename P3,
               typename T0, typename T1, typename T2, typename T3>
      typename function_traits<Fn>::return_type
      call (Fn &fn,
            const tuple<P0, P1, P2, P3> &bounds,
            const tuple<T0, T1, T2, T3> &args) const
      {
        return fn ();
      }
    };

    template<>
    struct binder_call<1>
    {
      template<typename Fn,
               typename P0, typename P1, typename P2, typename P3,
               typename T0, typename T1, typename T2, typename T3>
      typename function_traits<Fn>::return_type
      call (Fn &fn,
            const tuple<P0, P1, P2, P3> &bounds,
            const tuple<T0, T1, T2, T3> &args) const
      {
        return fn (select_value (nth_elt<0> (bounds), args));
      }
    };

    template<>
    struct binder_call<2>
    {
      template<typename Fn,
               typename P0, typename P1, typename P2, typename P3,
               typename T0, typename T1, typename T2, typename T3>
      typename function_traits<Fn>::return_type
      call (Fn &fn,
            const tuple<P0, P1, P2, P3> &bounds,
            const tuple<T0, T1, T2, T3> &args) const
      {
        return fn (select_value (nth_elt<0> (bounds), args),
                   select_value (nth_elt<1> (bounds), args));
      }
    };

    template<>
    struct binder_call<3>
    {
      template<typename Fn,
               typename P0, typename P1, typename P2, typename P3,
               typename T0, typename T1, typename T2, typename T3>
      typename function_traits<Fn>::return_type
      call (Fn &fn,
            const tuple<P0, P1, P2, P3> &bounds,
            const tuple<T0, T1, T2, T3> &args) const
      {
        return fn (select_value (nth_elt<0> (bounds), args),
                   select_value (nth_elt<1> (bounds), args),
                   select_value (nth_elt<2> (bounds), args));
      }
    };

    template<>
    struct binder_call<4>
    {
      template<typename Fn,
               typename P0, typename P1, typename P2, typename P3,
               typename T0, typename T1, typename T2, typename T3>
      typename function_traits<Fn>::return_type
      call (Fn &fn,
            const tuple<P0, P1, P2, P3> &bounds,
            const tuple<T0, T1, T2, T3> &args) const
      {
        return fn (select_value (nth_elt<0> (bounds), args),
                   select_value (nth_elt<1> (bounds), args),
                   select_value (nth_elt<2> (bounds), args),
                   select_value (nth_elt<3> (bounds), args));
      }
    };


    template<typename Tp, typename T>
    struct binder_helper_base
    {
      template<typename Fn>
      binder_helper_base (const Fn &fn, const T &bounds)
        : fn_ (fn), bounds_ (bounds)
      {}

      function<Tp> fn_;
      T bounds_;
    };


    template<unsigned int N, typename Tp, typename P> struct binder_helper;

    template<typename Tp, typename P>
    struct binder_helper<0, Tp, P>
      : public binder_helper_base<Tp, P>
    {
      template<typename Fn>
      binder_helper (const Fn &fn, const P &bounds)
        : binder_helper_base<Tp, P> (fn, bounds)
      {}

      typename function_traits<Tp>::return_type
      operator() ()
      {
        binder_call<function_traits<Tp>::arity> b;

        return b.call (this->fn_, this->bounds_, make_tuple ());
      }
    };

    template<typename Tp, typename P>
    struct binder_helper<1, Tp, P>
      : public binder_helper_base<Tp, P>
    {
      template<typename Fn>
      binder_helper (const Fn &fn, const P &bounds)
        : binder_helper_base<Tp, P> (fn, bounds)
      {}

      template<typename T0>
      typename function_traits<Tp>::return_type
      operator() (const T0 &t0)
      {
        binder_call<function_traits<Tp>::arity> b;

        return b.call (this->fn_, this->bounds_, make_tuple (t0));
      }
    };

    template<typename Tp, typename P>
    struct binder_helper<2, Tp, P>
      : public binder_helper_base<Tp, P>
    {
      template<typename Fn>
      binder_helper (const Fn &fn, const P &bounds)
        : binder_helper_base<Tp, P> (fn, bounds)
      {}

      template<typename T0, typename T1>
      typename function_traits<Tp>::return_type
      operator() (const T0 &t0, const T1 &t1)
      {
        binder_call<function_traits<Tp>::arity> b;

        return b.call (this->fn_, this->bounds_, make_tuple (t0, t1));
      }
    };

    template<typename Tp, typename P>
    struct binder_helper<3, Tp, P>
      : public binder_helper_base<Tp, P>
    {
      template<typename Fn>
      binder_helper (const Fn &fn, const P &bounds)
        : binder_helper_base<Tp, P> (fn, bounds)
      {}

      template<typename T0, typename T1, typename T2>
      typename function_traits<Tp>::return_type
      operator() (const T0 &t0, const T1 &t1, const T2 &t2)
      {
        binder_call<function_traits<Tp>::arity> b;

        return b.call (this->fn_, this->bounds_, make_tuple (t0, t1, t2));
      }
    };

    template<typename Tp, typename P>
    struct binder_helper<4, Tp, P>
      : public binder_helper_base<Tp, P>
    {
      template<typename Fn>
      binder_helper (const Fn &fn, const P &bounds)
        : binder_helper_base<Tp, P> (fn, bounds)
      {}

      template<typename T0, typename T1, typename T2, typename T3>
      typename function_traits<Tp>::return_type
      operator() (const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3)
      {
        binder_call<function_traits<Tp>::arity> b;

        return b.call (this->fn_, this->bounds_, make_tuple (t0, t1, t2, t3));
      }
    };


    template<typename Tp, typename P>
    struct binder
      : public binder_helper<rank<typename P::cons_type>::value, Tp, P>
    {
      template<typename Fn>
      binder (const Fn &fn, const P &p)
        : binder_helper<rank<typename P::cons_type>::value, Tp, P> (fn, p)
      {}
    };
  }


  template<typename Fn, typename P0>
  inline impl::binder<typename function_traits<Fn>::type, tuple<P0> >
  bind (const Fn &fn, const P0 &p0)
  {
    return impl::binder<typename function_traits<Fn>::type, tuple<P0> >
      (fn, make_tuple (p0));
  }

  template<typename Fn, typename P0, typename P1>
  inline impl::binder<typename function_traits<Fn>::type, tuple<P0, P1> >
  bind (const Fn &fn, const P0 &p0, const P1 &p1)
  {
    return impl::binder<typename function_traits<Fn>::type, tuple<P0, P1> >
      (fn, make_tuple (p0, p1));
  }

  template<typename Fn, typename P0, typename P1, typename P2>
  inline impl::binder<typename function_traits<Fn>::type, tuple<P0, P1, P2> >
  bind (const Fn &fn, const P0 &p0, const P1 &p1, const P2 &p2)
  {
    return impl::binder<typename function_traits<Fn>::type, tuple<P0, P1, P2> >
      (fn, make_tuple (p0, p1, p2));
  }

  template<typename Fn, typename P0, typename P1, typename P2, typename P3>
  inline impl::binder<typename function_traits<Fn>::type, tuple<P0, P1, P2, P3> >
  bind (const Fn &fn, const P0 &p0, const P1 &p1, const P2 &p2, const P3 &p3)
  {
    return impl::binder<typename function_traits<Fn>::type, tuple<P0, P1, P2, P3> >
      (fn, make_tuple (p0, p1, p2, p3));
  }

}

#endif /* nonstd__bind_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: a5e7a788-8b10-49fe-9fcd-e765d4cdc6f8
 * End:
 */
