/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__function_traits_h
#define nonstd__function_traits_h 1

#include "misc.h"

namespace nonstd
{
  template<typename T>
  struct function_traits : public T::traits {};

  template<typename R>
  struct function_traits<R (void)>
  {
    static const unsigned int arity = 0;
    typedef R type (void);
    typedef R return_type;
    typedef impl::none arg0_type;
    typedef impl::none arg1_type;
    typedef impl::none arg2_type;
    typedef impl::none arg3_type;
  };

  template<typename R>
  struct function_traits<R (*) (void)>
    : public function_traits<R (void)>
  {};

  template<typename T, typename R>
  struct function_traits<R (T::*) (void)>
    : public function_traits<R (void)>
  {};


  template<typename R, typename Arg0>
  struct function_traits<R (Arg0)>
  {
    static const unsigned int arity = 1;
    typedef R type (Arg0);
    typedef R return_type;
    typedef Arg0 arg0_type;
    typedef impl::none arg1_type;
    typedef impl::none arg2_type;
    typedef impl::none arg3_type;
  };

  template<typename R, typename Arg0>
  struct function_traits<R (*) (Arg0)>
    : public function_traits<R (Arg0)>
  {};

  template<typename T, typename R, typename Arg0>
  struct function_traits<R (T::*) (Arg0)>
    : public function_traits<R (Arg0)>
  {};


  template<typename R, typename Arg0, typename Arg1>
  struct function_traits<R (Arg0, Arg1)>
  {
    static const unsigned int arity = 2;
    typedef R type (Arg0, Arg1);
    typedef R return_type;
    typedef Arg0 arg0_type;
    typedef Arg1 arg1_type;
    typedef impl::none arg2_type;
    typedef impl::none arg3_type;
  };

  template<typename R, typename Arg0, typename Arg1>
  struct function_traits<R (*) (Arg0, Arg1)>
    : public function_traits<R (Arg0, Arg1)>
  {};

  template<typename T, typename R, typename Arg0, typename Arg1>
  struct function_traits<R (T::*) (Arg0, Arg1)>
    : public function_traits<R (Arg0, Arg1)>
  {};


  template<typename R, typename Arg0, typename Arg1, typename Arg2>
  struct function_traits<R (Arg0, Arg1, Arg2)>
  {
    static const unsigned int arity = 3;
    typedef R type (Arg0, Arg1, Arg2);
    typedef R return_type;
    typedef Arg0 arg0_type;
    typedef Arg1 arg1_type;
    typedef Arg2 arg2_type;
    typedef impl::none arg3_type;
  };

  template<typename R, typename Arg0, typename Arg1, typename Arg2>
  struct function_traits<R (*) (Arg0, Arg1, Arg2)>
    : public function_traits<R (Arg0, Arg1, Arg2)>
  {};

  template<typename T, typename R, typename Arg0, typename Arg1, typename Arg2>
  struct function_traits<R (T::*) (Arg0, Arg1, Arg2)>
    : public function_traits<R (Arg0, Arg1, Arg2)>
  {};


  template<typename R, typename Arg0, typename Arg1, typename Arg2,
           typename Arg3>
  struct function_traits<R (Arg0, Arg1, Arg2, Arg3)>
  {
    static const unsigned int arity = 4;
    typedef R type (Arg0, Arg1, Arg2, Arg3);
    typedef R return_type;
    typedef Arg0 arg0_type;
    typedef Arg1 arg1_type;
    typedef Arg2 arg2_type;
    typedef Arg3 arg3_type;
  };

  template<typename R, typename Arg0, typename Arg1, typename Arg2,
           typename Arg3>
  struct function_traits<R (*) (Arg0, Arg1, Arg2, Arg3)>
    : public function_traits<R (Arg0, Arg1, Arg2, Arg3)>
  {};

  template<typename T, typename R, typename Arg0, typename Arg1, typename Arg2,
           typename Arg3>
  struct function_traits<R (T::*) (Arg0, Arg1, Arg2, Arg3)>
    : public function_traits<R (Arg0, Arg1, Arg2, Arg3)>
  {};
}

#endif /* nonstd__function_traits_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: 61d5a60b-96d7-428d-91eb-29db218c9ea3
 * End:
 */
