/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__function_h
#define nonstd__function_h 1

#include "type-traits.h"
#include "function-traits.h"
#include "ptr-to-member.h"

#include <memory>

namespace nonstd
{
  namespace impl
  {
    template<unsigned int arity, typename Fn> struct function_base;

    template<typename Tp>
    struct function_base<0, Tp>
    {
      virtual ~function_base () {}

      virtual function_base<0, Tp> *clone () const = 0;

      virtual typename function_traits<Tp>::return_type operator() () = 0;
    };

    template<typename Tp>
    struct function_base<1, Tp>
    {
      virtual ~function_base () {}

      virtual function_base<1, Tp> *clone () const = 0;

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type) = 0;
    };

    template<typename Tp>
    struct function_base<2, Tp>
    {
      virtual ~function_base () {}

      virtual function_base<2, Tp> *clone () const = 0;

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type,
                  typename function_traits<Tp>::arg1_type) = 0;
    };

    template<typename Tp>
    struct function_base<3, Tp>
    {
      virtual ~function_base () {}

      virtual function_base<3, Tp> *clone () const = 0;

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type,
                  typename function_traits<Tp>::arg1_type,
                  typename function_traits<Tp>::arg2_type) = 0;
    };

    template<typename Tp>
    struct function_base<4, Tp>
    {
      virtual ~function_base () {}

      virtual function_base<4, Tp> *clone () const = 0;

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type,
                  typename function_traits<Tp>::arg1_type,
                  typename function_traits<Tp>::arg2_type,
                  typename function_traits<Tp>::arg3_type ) = 0;
    };


    template<unsigned int arity, typename Tp, class Fn>
    struct function_holder;

    template<typename Tp, typename Fn>
    struct function_holder<0, Tp, Fn> : public function_base<0, Tp>
    {
      function_holder (const Fn &fn) : fn_ (fn) {}

      virtual function_base<0, Tp> *
      clone () const
      {
        return new function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() ()
      {
        return fn_ ();
      }

      Fn fn_;
    };

    template<typename Tp, typename Fn>
    struct function_holder<1, Tp, Fn> : public function_base<1, Tp>
    {
      function_holder (const Fn &fn) : fn_ (fn) {}

      virtual function_base<1, Tp> *
      clone () const
      {
        return new function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type arg0)
      {
        return fn_ (arg0);
      }

      Fn fn_;
    };

    template<typename Tp, typename Fn>
    struct function_holder<2, Tp, Fn> : public function_base<2, Tp>
    {
      function_holder (const Fn &fn) : fn_ (fn) {}

      virtual function_base<2, Tp> *
      clone () const
      {
        return new function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1)
      {
        return fn_ (arg0, arg1);
      }

      Fn fn_;
    };

    template<typename Tp, typename Fn>
    struct function_holder<3, Tp, Fn> : public function_base<3, Tp>
    {
      function_holder (const Fn &fn) : fn_ (fn) {}

      virtual function_base<3, Tp> *
      clone () const
      {
        return new function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1,
                  typename function_traits<Tp>::arg2_type arg2)
      {
        return fn_ (arg0, arg1, arg2);
      }

      Fn fn_;
    };

    template<typename Tp, typename Fn>
    struct function_holder<4, Tp, Fn> : public function_base<4, Tp>
    {
      function_holder (const Fn &fn) : fn_ (fn) {}

      virtual function_base<4, Tp> *
      clone () const
      {
        return new function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1,
                  typename function_traits<Tp>::arg2_type arg2,
                  typename function_traits<Tp>::arg3_type arg3)
      {
        return fn_ (arg0, arg1, arg2, arg3);
      }

      Fn fn_;
    };


    template<unsigned int arity, class Obj, typename Tp>
    struct mem_function_holder;

    template<class Obj, typename Tp>
    struct mem_function_holder<0, Obj, Tp>
      : public function_base<1, typename prepend_arg<Obj *, Tp>::type>
    {
      mem_function_holder (typename ptr_to_member<Obj, Tp>::type fn)
        : fn_ (fn)
      {
      }

      virtual function_base<1, typename prepend_arg<Obj *, Tp>::type> *
      clone () const
      {
        return new mem_function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (Obj *obj)
      {
        return (obj->*fn_) ();
      }

      typename ptr_to_member<Obj, Tp>::type fn_;
    };

    template<class Obj, typename Tp>
    struct mem_function_holder<1, Obj, Tp>
      : public function_base<2, typename prepend_arg<Obj *, Tp>::type>
    {
      mem_function_holder (typename ptr_to_member<Obj, Tp>::type fn)
        : fn_ (fn)
      {
      }

      virtual function_base<2, typename prepend_arg<Obj *, Tp>::type> *
      clone () const
      {
        return new mem_function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (Obj *obj,
                  typename function_traits<Tp>::arg0_type arg0)
      {
        return (obj->*fn_) (arg0);
      }

      typename ptr_to_member<Obj, Tp>::type fn_;
    };

    template<class Obj, typename Tp>
    struct mem_function_holder<2, Obj, Tp>
      : public function_base<3, typename prepend_arg<Obj *, Tp>::type>
    {
      mem_function_holder (typename ptr_to_member<Obj, Tp>::type fn)
        : fn_ (fn)
      {
      }

      virtual function_base<3, typename prepend_arg<Obj *, Tp>::type> *
      clone () const
      {
        return new mem_function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (Obj *obj,
                  typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1)
      {
        return (obj->*fn_) (arg0, arg1);
      }

      typename ptr_to_member<Obj, Tp>::type fn_;
    };

    template<class Obj, typename Tp>
    struct mem_function_holder<3, Obj, Tp>
      : public function_base<4, typename prepend_arg<Obj *, Tp>::type>
    {
      mem_function_holder (typename ptr_to_member<Obj, Tp>::type fn)
        : fn_ (fn)
      {
      }

      virtual function_base<4, typename prepend_arg<Obj *, Tp>::type> *
      clone () const
      {
        return new mem_function_holder (fn_);
      }

      virtual typename function_traits<Tp>::return_type
      operator() (Obj *obj,
                  typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1,
                  typename function_traits<Tp>::arg2_type arg2)
      {
        return (obj->*fn_) (arg0, arg1, arg2);
      }

      typename ptr_to_member<Obj, Tp>::type fn_;
    };
  }


  template<typename Tp>
  class function
  {
  public:
    typedef function_traits<Tp> traits;

    function (const function<Tp> &fn)
      : fn_ (fn.fn_->clone ())
    {}

    template<typename Fn>
    function (const Fn &fn)
      : fn_ (new impl::function_holder<function_traits<Tp>::arity, Tp, Fn> (fn))
    {}

    template<typename Obj, typename R>
    function (R (Obj::*fn) (void))
      : fn_ (new impl::mem_function_holder<0, Obj, R (void)> (fn))
    {}

    template<typename Obj, typename R, typename A0>
    function (R (Obj::*fn) (A0))
      : fn_ (new impl::mem_function_holder<1, Obj, R (A0)> (fn))
    {}

    template<typename Obj, typename R, typename A0, typename A1>
    function (R (Obj::*fn) (A0, A1))
      : fn_ (new impl::mem_function_holder<2, Obj, R (A0, A1)> (fn))
    {
    }

    template<typename Obj, typename R, typename A0, typename A1, typename A2>
    function (R (Obj::*fn) (A0, A1, A2))
      : fn_ (new impl::mem_function_holder<3, Obj, R (A0, A1, A2)> (fn))
    {
    }

    typename function_traits<Tp>::return_type
    operator() ()
    {
      return (*fn_.get ()) ();
    };

    typename function_traits<Tp>::return_type
    operator() (typename function_traits<Tp>::arg0_type arg0)
    {
      return (*fn_.get ()) (arg0);
    }

    typename function_traits<Tp>::return_type
    operator() (typename function_traits<Tp>::arg0_type arg0,
                typename function_traits<Tp>::arg1_type arg1)
    {
      return (*fn_.get ()) (arg0, arg1);
    }

    typename function_traits<Tp>::return_type
    operator() (typename function_traits<Tp>::arg0_type arg0,
                typename function_traits<Tp>::arg1_type arg1,
                typename function_traits<Tp>::arg2_type arg2)
    {
      return (*fn_.get ()) (arg0, arg1, arg2);
    }

    typename function_traits<Tp>::return_type
    operator() (typename function_traits<Tp>::arg0_type arg0,
                typename function_traits<Tp>::arg1_type arg1,
                typename function_traits<Tp>::arg2_type arg2,
                typename function_traits<Tp>::arg3_type arg3)
    {
      return (*fn_.get ()) (arg0, arg1, arg2, arg3);
    }

  protected:
    std::auto_ptr<impl::function_base<function_traits<Tp>::arity, Tp> > fn_;
  };


  template<typename Fn>
  inline function<typename function_traits<Fn>::type>
  make_function (const Fn &fn)
  {
    return function<typename function_traits<Fn>::type> (fn);
  }


  template<typename Obj, typename R>
  inline function<R (Obj *)>
  make_function (R (Obj::*fn) (void))
  {
    return function<R (Obj *)> (fn);
  }

  template<typename Obj, typename R, typename A0>
  inline function<R (Obj *, A0)>
  make_function (R (Obj::*fn) (A0))
  {
    return function<R (Obj *, A0)> (fn);
  }

  template<typename Obj, typename R, typename A0, typename A1>
  inline function<R (Obj *, A0, A1)>
  make_function (R (Obj::*fn) (A0, A1))
  {
    return function<R (Obj *, A0, A1)> (fn);
  }

  template<typename Obj, typename R, typename A0, typename A1, typename A2>
  inline function<R (Obj *, A0, A1, A2)>
  make_function (R (Obj::*fn) (A0, A1, A2))
  {
    return function<R (Obj *, A0, A1, A2)> (fn);
  }
}

#endif /* nonstd__function_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: 61377eae-5d40-42dd-9341-8beb4cc915f5
 * End:
 */
