#include "smart-ptr.h"
#include <iostream>

using namespace std;
using namespace nonstd;

namespace
{
  class integer
  {
  public:
    explicit integer (int i = 0)
      : value_ (i), count_ (1)
    {
      cout << "integer::integer (" << i << ")" << endl;
    }

    integer (const integer &i)
      : value_ (i.value_)
    {
      cout << "integer::integer (const integer &)" << endl;
    }

    ~integer ()
    {
      cout << "integer::~integer ()" << endl;
    }

    integer &
    operator= (const integer &i)
    {
      cout << "integer::operator= (const integer &)" << endl;
      value_ = i.value_;
      return *this;
    }

    integer &
    operator= (int i)
    {
      cout << "integer::operator= (int)" << endl;
      value_ = i;
      return *this;
    }

    operator const int& () const
    {
      return value_;
    }

    integer *
    duplicate ()
    {
      cout << "integer::duplicate ()" << endl;
      count_++;
      return this;
    }

    bool
    release ()
    {
      cout << "integer::release ()" << endl;
      return --count_ == 0;
    }

  protected:
    int value_;
    unsigned int count_;
  };
}

int
main ()
{
  cout << "*** Non-intrusive ref count" << endl;
  {
    typedef smart_ptr<integer> sp;

    sp p (new integer (1));
    cout << *p << endl;

    {
      sp q (p);
      *q = 2;
    }

    cout << *p << endl;

    sp q;
    q = p;
    *q = 3;

    cout << *p << endl;

    *q = integer (4);

    cout << *p << endl;

    q = new integer (5);

    cout << *p << endl;
  }

  cout << endl << "*** Intrusive ref count" << endl;
  {
    typedef smart_ptr<integer, default_storage_policy, intrusive_ref_count> sp;

    sp p (new integer (1));
    cout << *p << endl;

    {
      sp q (p);
      *q = 2;
    }

    cout << *p << endl;

    sp q;
    q = p;
    *q = 3;

    cout << *p << endl;

    *q = integer (4);

    cout << *p << endl;

    q = new integer (5);

    cout << *p << endl;
  }

  return 0;
}

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: 1b72f85f-f28b-4caa-a36e-7be539a4cc55
 * End:
 */
