/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__tuple_h
#define nonstd__tuple_h 1

#include "call-traits.h"
#include "function-traits.h"

namespace nonstd
{
  namespace impl
  {
    template<typename H, typename T> struct cons;

    template<>
    struct cons<nil, nil>
    {
      cons (const nil &, const nil &, const nil &, const nil &)
      {}
    };

    template<typename H>
    struct cons<H, nil>
    {
      H head;

      cons (const H &h = wrap<H> ())
        : head (h)
      {}

      cons (const H &h, const nil &, const nil &, const nil &)
        : head (h)
      {}

      template<typename H1>
      cons (const cons<H1, nil> &c)
        : head (c.head)
      {}
    };

    template<typename H, typename T>
    struct cons
    {
      H head;
      T tail;

      template<typename T1, typename T2, typename T3>
      cons (const H &t0, const T1 &t1, const T2 &t2, const T3 &t3)
        : head (t0), tail (t1, t2, t3, nil ())
      {}

      template<typename H1, typename T1>
      cons (const cons<H1, T1> &c)
        : head (c.head), tail (c.tail)
      {}
    };


    template<unsigned int N> struct nth_elt_helper;

    template<>
    struct nth_elt_helper<0>
    {
      template<typename H, typename T>
      static void *get (cons<H, T> &c)
      {
        return &c.head;
      }

      template<typename H, typename T>
      static const void *get (const cons<H, T> &c)
      {
        return &c.head;
      }
    };

    template<unsigned int N>
    struct nth_elt_helper
    {
      template<typename H, typename T>
      static void *get (cons<H, T> &c)
      {
        return nth_elt_helper<N-1>::get (c.tail);
      }

      template<typename H, typename T>
      static const void *get (const cons<H, T> &c)
      {
        return nth_elt_helper<N-1>::get (c.tail);
      }
    };


    template<typename T0, typename T1, typename T2, typename T3>
    struct tuple_to_cons;

    template<>
    struct tuple_to_cons<nil, nil, nil, nil>
    {
      typedef cons<nil, nil> type;
    };

    template<typename T0, typename T1, typename T2, typename T3>
    struct tuple_to_cons
    {
      typedef cons<T0, typename tuple_to_cons<T1, T2, T3, nil>::type> type;
    };

  }

  template<typename T0, typename T1 = impl::nil, typename T2 = impl::nil,
           typename T3 = impl::nil>
  struct tuple
    : public impl::tuple_to_cons<T0, T1, T2, T3>::type
  {
    typedef typename impl::tuple_to_cons<T0, T1, T2, T3>::type cons_type;

    tuple (typename param<T0>::type t0 = impl::wrap<T0> (),
           typename param<T1>::type t1 = impl::wrap<T1> (),
           typename param<T2>::type t2 = impl::wrap<T2> (),
           typename param<T3>::type t3 = impl::wrap<T3> ())
      : impl::tuple_to_cons<T0, T1, T2, T3>::type (t0, t1, t2, t3)
    {}

    template<typename H, typename T>
    tuple (const impl::cons<H, T> &c)
      : impl::tuple_to_cons<T0, T1, T2, T3> (c)
    {}
  };


  inline tuple<impl::nil>
  make_tuple ()
  {
    return tuple<impl::nil> ();
  }

  template<typename T0>
  inline tuple<T0>
  make_tuple (const T0 &t0)
  {
    return tuple<T0> (t0);
  }

  template<typename T0, typename T1>
  inline tuple<T0, T1>
  make_tuple (const T0 &t0, const T1 &t1)
  {
    return tuple<T0, T1> (t0, t1);
  }

  template<typename T0, typename T1, typename T2>
  inline tuple<T0, T1, T2>
  make_tuple (const T0 &t0, const T1 &t1, const T2 &t2)
  {
    return tuple<T0, T1, T2> (t0, t1, t2);
  }

  template<typename T0, typename T1, typename T2, typename T3>
  inline tuple<T0, T1, T2, T3>
  make_tuple (const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3)
  {
    return tuple<T0, T1, T2, T3> (t0, t1, t2, t3);
  }


  template<unsigned int N, typename T> struct nth_type;

  template<typename T0, typename T1, typename T2, typename T3>
  struct nth_type<0, tuple<T0, T1, T2, T3> >
  {
    typedef T0 type;
  };

  template<unsigned int N, typename T0, typename T1, typename T2, typename T3>
  struct nth_type<N, tuple<T0, T1, T2, T3> >
  {
    typedef typename nth_type<N - 1, tuple<T1, T2, T3> >::type type;
  };


  template<unsigned int N, typename T0, typename T1, typename T2, typename T3>
  inline const typename nth_type<N, tuple<T0, T1, T2, T3> >::type &
  nth_elt (const tuple<T0, T1, T2, T3> &t)
  {
    typedef const typename nth_type<N, tuple<T0, T1, T2, T3> >::type elt_type;

    return *static_cast<elt_type *> (impl::nth_elt_helper<N>::get (t));
  };

  template<unsigned int N, typename T0, typename T1, typename T2, typename T3>
  inline typename nth_type<N, tuple<T0, T1, T2, T3> >::type &
  nth_elt (tuple<T0, T1, T2, T3> &t)
  {
    typedef typename nth_type<N, tuple<T0, T1, T2, T3> >::type elt_type;

    return *static_cast<elt_type *> (impl::nth_elt_helper<N>::get (t));
  };


  template<typename Fn, typename T0>
  inline typename function_traits<Fn>::return_type
  call (Fn &fn, const tuple<T0> &t)
  {
    return fn (nth_elt<0> (t));
  }

  template<typename Fn, typename T0>
  inline typename function_traits<Fn>::return_type
  call (const Fn &fn, const tuple<T0> &t)
  {
    return fn (nth_elt<0> (t));
  }

  template<typename Fn, typename T0, typename T1>
  inline typename function_traits<Fn>::return_type
  call (Fn &fn, const tuple<T0, T1> &t)
  {
    return fn (nth_elt<0> (t), nth_elt<1> (t));
  }

  template<typename Fn, typename T0, typename T1>
  inline typename function_traits<Fn>::return_type
  call (const Fn &fn, const tuple<T0, T1> &t)
  {
    return fn (nth_elt<0> (t), nth_elt<1> (t));
  }

  template<typename Fn, typename T0, typename T1, typename T2>
  inline typename function_traits<Fn>::return_type
  call (Fn &fn, const tuple<T0, T1, T2> &t)
  {
    return fn (nth_elt<0> (t), nth_elt<1> (t), nth_elt<2> (t));
  }

  template<typename Fn, typename T0, typename T1, typename T2>
  inline typename function_traits<Fn>::return_type
  call (const Fn &fn, const tuple<T0, T1, T2> &t)
  {
    return fn (nth_elt<0> (t), nth_elt<1> (t), nth_elt<2> (t));
  }

  template<typename Fn, typename T0, typename T1, typename T2, typename T3>
  inline typename function_traits<Fn>::return_type
  call (const Fn &fn, const tuple<T0, T1, T2, T3> &t)
  {
    return fn (nth_elt<0> (t), nth_elt<1> (t), nth_elt<2> (t), nth_elt<3> (t));
  }
}

#endif /* nonstd__tuple_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: 63d3639d-35a5-433c-81db-04bf45f3293c
 * End:
 */
