/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__type_traits_h
#define nonstd__type_traits_h 1

namespace nonstd
{
  template<typename Tp>
  struct add_const
  {
    typedef Tp const type;
  };

  template<typename Tp>
  struct add_const<const Tp>
  {
    typedef Tp const type;
  };

  template<typename Tp>
  struct remove_const
  {
    typedef Tp type;
  };

  template<typename Tp>
  struct remove_const<const Tp>
  {
    typedef Tp type;
  };


  template<typename Tp>
  struct add_volatile
  {
    typedef Tp volatile type;
  };

  template<typename Tp>
  struct add_volatile<volatile Tp>
  {
    typedef Tp volatile type;
  };

  template<typename Tp>
  struct remove_volatile
  {
    typedef Tp type;
  };

  template<typename Tp>
  struct remove_volatile<volatile Tp>
  {
    typedef Tp type;
  };


  template<typename Tp>
  struct add_pointer
  {
    typedef Tp *type;
  };

  template<typename Tp>
  struct add_pointer<Tp &>
  {
    typedef Tp *type;
  };

  template<typename Tp>
  struct remove_pointer
  {
    typedef Tp type;
  };

  template<typename Tp>
  struct remove_pointer<Tp *>
  {
    typedef Tp type;
  };


  template<typename Tp>
  struct add_reference
  {
    typedef Tp &type;
  };

  template<typename Tp>
  struct add_reference<Tp &>
  {
    typedef Tp &type;
  };

  template<typename Tp>
  struct remove_reference
  {
    typedef Tp type;
  };

  template<typename Tp>
  struct remove_reference<Tp &>
  {
    typedef Tp type;
  };


  template<typename Tp>
  struct is_primitive
  {
    static const bool value = false;
  };

  template<>
  struct is_primitive<char>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<signed char>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<unsigned char>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<signed short>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<unsigned short>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<signed int>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<unsigned int>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<signed long>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<unsigned long>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<signed long long>
  {
    static const bool value = true;
  };

  template<>
  struct is_primitive<unsigned long long>
  {
    static const bool value = true;
  };


  template<typename Tp>
  struct is_pointer
  {
    static const bool value = false;
  };

  template<typename Tp>
  struct is_pointer<Tp *>
  {
    static const bool value = true;
  };


  template<typename Tp>
  struct is_reference
  {
    static const bool value = false;
  };

  template<typename Tp>
  struct is_reference<Tp &>
  {
    static const bool value = true;
  };


  template<typename A, typename Fn>  struct prepend_arg;

  template<typename A, typename R>
  struct prepend_arg<A, R ()>
  {
    typedef R (type) (A);
  };

  template<typename A, typename R, typename A0>
  struct prepend_arg<A, R (A0)>
  {
    typedef R (type) (A, A0);
  };

  template<typename A, typename R, typename A0, typename A1>
  struct prepend_arg<A, R (A0, A1)>
  {
    typedef R (type) (A, A0, A1);
  };

  template<typename A, typename R, typename A0, typename A1, typename A2>
  struct prepend_arg<A, R (A0, A1, A2)>
  {
    typedef R (type) (A, A0, A1, A2);
  };
}

#endif /* nonstd__type_traits_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: 9247b0f5-b083-4a8c-9d0e-a3d74e95c89e
 * End:
 */
